#!/usr/bin/perl

# This is a quick script to generate mpixl* from mpi{cc|cxx|f77} scripts and a single config file.
# Arguments are path-to-mpich2-install-dir
#
# It is run in the MPI-level makefile

use warnings;
use File::Copy;

($prefix_path = shift) or (die "Must provide location of existing compiler scripts\n");

(-e "$prefix_path/mpicc") or (die "Must provide location of existing compiler scripts\n");

@files = <$prefix_path/*>;
foreach $file (@files) {
   $file_tmp = "${file}_tmp";
   open SCRIPTFILE,   "<$file"    || die;
   open SCRIPTFILE_TMP,   ">$file_tmp"    || die;
   chFile(SCRIPTFILE,SCRIPTFILE_TMP);
   close SCRIPTFILE;
   close SCRIPTFILE_TMP;
   copy($file_tmp, $file);
   unlink($file_tmp);
} 

foreach $file (@files) {
   if($file =~ m/xl/) { 
   	$file_tmp = "${file}_tmp";
	open SCRIPTFILE,   "<$file"    || die;
	open SCRIPTFILE_TMP,   ">$file_tmp"    || die;
   	chFile2(SCRIPTFILE,SCRIPTFILE_TMP);
	close SCRIPTFILE;
	close SCRIPTFILE_TMP;
	copy($file_tmp, $file);
	unlink($file_tmp);
   } 
}

sub chFile
{
	$ifilename = $_[0];
	$filename = $_[1];
	while(<$ifilename>)
	{ 
    		if (/^prefix=/) 
    		{ 
			if ( ($pos = index($prefix_path, 'pe')) >= 0) {
			        $release_nm = substr $prefix_path, $pos, 6;
			}
			if ($prefix_path =~ m/intel/) { 
        			print $filename  "prefix=/opt/ibmhpc/${release_nm}/mpich2/intel\n"; 
			} 
			else {
        			print $filename  "prefix=/opt/ibmhpc/${release_nm}/mpich2/gnu\n"; 
			} 
    		} 
    		elsif (/^includedir=/) { 
			if( ($prefix_path =~ m/64/) || ($prefix_path =~ m/intel/) ) {	 
       				print $filename  "includedir=\$\{prefix\}/include64\n"; 
			} 
			else { 
       				print $filename  $_; 
			}
    		} 
		elsif (/^libdir=/) { 
			if( ($prefix_path =~ m/64/) || ($prefix_path =~ m/intel/) ) {	 
       				print $filename "libdir=\$\{prefix\}/lib64\n"; 
			} 
			else { 
       				print $filename  $_; 
			}
    		} 
    		elsif (/^CFLAGS=/) { 
			if ($prefix_path =~ m/intel/) { 
				print $filename "CFLAGS=\"\"\n"; 
			} 
			elsif ($prefix_path =~ m/64/) { 
				print $filename "CFLAGS=\" -m64 \"\n"; 
			} 
			else {
				print $filename "CFLAGS=\" -m32 \"\n"; 
			}
    		} 
    		elsif (/^CXXFLAGS=/) { 
			if ($prefix_path =~ m/intel/) { 
				print $filename "CXXFLAGS=\"\"\n"; 
			} 
			elsif ($prefix_path =~ m/64/) { 
				print $filename "CXXFLAGS=\" -m64 \"\n"; 
			} 
			else {
				print $filename "CXXFLAGS=\" -m32 \"\n"; 
			} 
    		} 
    		elsif (/^FFLAGS=/) { 
			if ($prefix_path =~ m/intel/) { 
				print $filename "FFLAGS=\"\"\n"; 
			} 
			elsif ($prefix_path =~ m/64/) { 
				print $filename "FFLAGS=\" -m64 \"\n"; 
			} 
			else {
				print $filename "FFLAGS=\" -m32 \"\n"; 
			} 
    		} 
    		elsif (/^FCFLAGS=/) { 
			if ($prefix_path =~ m/intel/) { 
				print $filename "FCFLAGS=\"\"\n"; 
			} 
			elsif ($prefix_path =~ m/64/) { 
				print $filename "FCFLAGS=\" -m64 \"\n"; 
			} 
			else {
				print $filename "FCFLAGS=\" -m32 \"\n"; 
			} 
    		} 
    		elsif (/^LIBS=/) { 
			if ($prefix_path =~ m/intel/) { 
				print $filename "LIBS=\"-lopa -lmpl -lpami -lpthread -lrt -lsvml -lirc\"\n"; 
			} 
			else {
				print $filename "LIBS=\"-lopa -lmpl -ldl -lpami -lpthread -lrt \"\n"; 
			}
    		} 
    		elsif (/^LDFLAGS=/) { 
			if ($prefix_path =~ m/intel/) { 
				print $filename "LDFLAGS=\"  -Wl,-melf_x86_64 -O3\"\n"; 
			} 
			else { 
       				print $filename  $_; 
			}
    		} 
    		elsif (/^CXX=/) { 
			if ($prefix_path =~ m/intel/) { 
				print $filename "CXX=\"icpc\"\n"; 
			} 
			else { 
       				print $filename  $_; 
			}
    		} 
    		else { 
     			print $filename  $_; 
    		} 
	}
}

sub chFile2
{
	$ifilename = $_[0];
	$filename = $_[1];
	while(<$ifilename>)
	{ 
    		if (/^CFLAGS=/) { 
			if ($prefix_path =~ m/64/) {  
				print $filename "CFLAGS=\"-q64\"\n"; 
			} 
			else { 
       				print $filename "CFLAGS=\"-q32\"\n" 
			}
    		} 
    		elsif (/^CXXFLAGS=/) { 
			if ($prefix_path =~ m/64/) {  
				print $filename "CXXFLAGS=\"-q64\"\n"; 
			} 
			else { 
       				print $filename "CXXFLAGS=\"-q32\"\n" 
			}
    		} 
    		elsif (/^LDFLAGS=/) { 
			if ($prefix_path =~ m/64/) {  
				print $filename "LDFLAGS=\"  -q64 -O3\"\n"; 
			} 
			else { 
       				print $filename "LDFLAGS=\"  -q32 -O3\"\n"; 
			}
    		} 
    		elsif (/^FFLAGS=/) { 
			if ($prefix_path =~ m/64/) {  
				print $filename "FFLAGS=\"-q64\"\n"; 
			} 
			else { 
       				print $filename "FFLAGS=\"-q32\"\n" 
			}
    		} 
    		else { 
     			print $filename  $_; 
    		} 
	}
}
