/* IBM_PROLOG_BEGIN_TAG                                                   */
/* This is an automatically generated prolog.                             */
/*                                                                        */
/*                                                                        */
/*                                                                        */
/* Licensed Materials - Property of IBM                                   */
/*                                                                        */
/* (C) COPYRIGHT International Business Machines Corp. 2010,2013          */
/* All Rights Reserved                                                    */
/*                                                                        */
/* US Government Users Restricted Rights - Use, duplication or            */
/* disclosure restricted by GSA ADP Schedule Contract with IBM Corp.      */
/*                                                                        */
/* IBM_PROLOG_END_TAG                                                     */
/* "@(#) 1.14 src/ppe/openshmem/include/shmem.h, ppe.openshmem, ppe_rbarlx2, rbarlx2s012a 13/03/14 14:13:01" */

#ifndef _SHMEM_H_
#define _SHMEM_H_
#include <stdlib.h>
#include <sys/types.h>

#ifdef __cplusplus
#include <cstddef>
#include </usr/include/complex.h> 
#else
#include <stddef.h>
#include <complex.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif 

#define SHMEM_ERR_NOERR        (0)    /* no error */
#define SHMEM_ERR_UNINTIALIZED (-101) /* not initialized */
#define SHMEM_ERR_INTIALIZED   (-102) /* already initialized */
#define SHMEM_ERR_HEAP_INIT    (-103) /* Symmetric heap failed to initialize */
#define SHMEM_ERR_HEAP_SPACE   (-104) /* Out of symm. heap space */
#define SHMEM_ERR_PARAMS       (-105) /* invalid parameters */
#define SHMEM_ERR_UNLOCKED     (-106) /* current pe tries to clean an empty lock */
#define SHMEM_ERR_LOCK_OWNER   (-107) /* current pe does not own the lock */        
#define SHMEM_ERR_COMM_SYS     (-108) /* error in underlying communication system */
#define SHMEM_ERR_SYS          (-109) /* system error; check errno*/
#define SHMEM_ERR              (-200) /* General error */
#define SHMEM_SUCCESS          SHMEM_ERR_NOERR

extern int  shmem_err;
extern long malloc_error;

#define _SHMEM_REDUCE_SYNC_SIZE        1
#define _SHMEM_REDUCE_MIN_WRKDATA_SIZE 1
#define _SHMEM_BARRIER_SYNC_SIZE       1
#define _SHMEM_BCAST_SYNC_SIZE         1
#define _SHMEM_COLLECT_SYNC_SIZE       1
#define _SHMEM_SYNC_VALUE              1
#define  SHMEM_CMP_EQ                  1
#define  SHMEM_CMP_NE                  2
#define  SHMEM_CMP_GT                  3
#define  SHMEM_CMP_LE                  4
#define  SHMEM_CMP_LT                  5
#define  SHMEM_CMP_GE                  6

#define MEM_INSUFFICIENT      0x40000011
#define MEM_ADDR_INVALID      0x40000222
#define MEM_ALIGN_INVALID     0x40003333

/* Initialization Routines */
void start_pes(int npes);
#define shmem_init() start_pes(0)
int _num_pes(void);
int _my_pe(void);
#define shmem_my_pe _my_pe
#define shmem_n_pes _num_pes
void shmem_finalize(void);

/* Accessibility Query Routines */
int shmem_pe_accessible(int pe);
int shmem_addr_accessible(void *addr, int pe);

/* Symmetric Heap Routines */
void *shmalloc(size_t size);
void *shmemalign(size_t alignment, size_t size);
void *shrealloc(void *ptr, size_t size);
void shfree(void *ptr);

/* Pointer Routine */
void *shmem_ptr(void *target, int pe);

/* Elemental Put Routines */
void shmem_short_p(short *addr, short value, int pe);
void shmem_int_p(int *addr, int value, int pe);
void shmem_long_p(long *addr, long value, int pe);
void shmem_float_p(float *addr, float value, int pe);
void shmem_double_p(double *addr, double value, int pe);
void shmem_longlong_p(long long *addr, long long value, int pe);
void shmem_longdouble_p(long double *addr, long double value, int pe);

/* Block Data Put Routines */
void shmem_char_put(char *target, const char *source, size_t nelems, int pe);
void shmem_short_put(short *target, const short *source, size_t nelems, int pe);
void shmem_int_put(int *target, const int *source, size_t nelems, int pe);
void shmem_long_put(long *target, const long *source, size_t nelems, int pe);
void shmem_float_put(float *target, const float *source, size_t nelems, int pe);
void shmem_double_put(double *target, const double *source, size_t nelems, int pe);
void shmem_longlong_put(long long *target, const long long *source, size_t nelems, int pe);
void shmem_longdouble_put(long double *target, const long double *source, size_t nelems, int pe);
void shmem_put32(void *target, const void *source, size_t nelems, int pe);
void shmem_put64(void *target, const void *source, size_t nelems, int pe);
void shmem_put128(void *target, const void *source, size_t nelems, int pe);
void shmem_putmem(void *target, const void *source, size_t nelems, int pe);

/* Strided Put Routines */
void shmem_short_iput(short *target, const short *source, ptrdiff_t tst, ptrdiff_t sst, size_t nelems,int pe);
void shmem_int_iput(int *target, const int *source, ptrdiff_t tst, ptrdiff_t sst, size_t nelems, int pe);
void shmem_float_iput(float *target, const float *source, ptrdiff_t tst, ptrdiff_t sst, size_t nelems,int pe);
void shmem_long_iput(long *target, const long *source, ptrdiff_t tst, ptrdiff_t sst, size_t nelems, int pe);
void shmem_double_iput(double *target, const double *source,ptrdiff_t tst, ptrdiff_t sst,size_t nelems,int pe);
void shmem_longlong_iput(long long *target, const long long *source, ptrdiff_t tst, ptrdiff_t sst, size_t nelems, int pe);
void shmem_longdouble_iput(long double *target, const long double *source, ptrdiff_t tst, ptrdiff_t sst, size_t nelems, int pe);
void shmem_iput32(void *target, const void *source, ptrdiff_t tst, ptrdiff_t sst, size_t nelems, int pe);
void shmem_iput64(void *target, const void *source, ptrdiff_t tst, ptrdiff_t sst, size_t nelems, int pe);
void shmem_iput128(void *target, const void *source, ptrdiff_t tst, ptrdiff_t sst, size_t nelems, int pe);

/* Elemental Get Routines */
short shmem_short_g(short *addr, int pe);
int shmem_int_g(int *addr, int pe);
long shmem_long_g(long *addr, int pe);
float shmem_float_g(float *addr, int pe);
double shmem_double_g(double *addr, int pe);
long long shmem_longlong_g(long long *addr, int pe);
long double shmem_longdouble_g(long double *addr, int pe);

/* Block Data Get Routines */
void shmem_char_get(char *target, const char *source, size_t nelems, int pe);
void shmem_short_get(short *target, const short *source, size_t nelems, int pe);
void shmem_int_get(int *target, const int *source, size_t nelems, int pe);
void shmem_long_get(long *target, const long *source, size_t nelems, int pe);
void shmem_float_get(float *target, const float *source, size_t nelems, int pe);
void shmem_double_get(double *target, const double *source, size_t nelems, int pe);
void shmem_longlong_get(long long *target, const long long *source, size_t nelems, int pe);
void shmem_longdouble_get(long double *target, const long double *source, size_t nelems, int pe);
void shmem_get32(void *target, const void *source, size_t nelems, int pe);
void shmem_get64(void *target, const void *source, size_t nelems, int pe);
void shmem_get128(void *target, const void *source, size_t nelems, int pe);
void shmem_getmem(void *target, const void *source, size_t nelems, int pe);

/* Strided Get Routines */ 
void shmem_iget32(void *target, const void *source, ptrdiff_t tst, ptrdiff_t sst, size_t nelems, int pe);
void shmem_iget64(void *target, const void *source, ptrdiff_t tst, ptrdiff_t sst, size_t nelems, int pe);
void shmem_iget128(void *target, const void *source, ptrdiff_t tst, ptrdiff_t sst, size_t nelems, int pe);
void shmem_short_iget(short *target, const short *source, ptrdiff_t tst, ptrdiff_t sst, size_t nelems,int pe);
void shmem_int_iget(int *target, const int *source, ptrdiff_t tst, ptrdiff_t sst, size_t nelems, int pe);
void shmem_long_iget(long *target, const long *source, ptrdiff_t tst, ptrdiff_t sst, size_t nelems, int pe);
void shmem_float_iget(float *target, const float *source, ptrdiff_t tst, ptrdiff_t sst, size_t nelems,int pe);
void shmem_double_iget(double *target, const double *source,ptrdiff_t tst, ptrdiff_t sst,size_t nelems,int pe);
void shmem_longlong_iget(long long *target, const long long *source, ptrdiff_t tst, ptrdiff_t sst, size_t nelems, int pe);
void shmem_longdouble_iget(long double *target, const long double *source, ptrdiff_t tst, ptrdiff_t sst, size_t nelems, int pe);

/* Atomic Memory fetch-and-operate Routines */
int shmem_int_swap(int *target, int value, int pe);
long shmem_long_swap(long *target, long value, int pe);
long shmem_swap(long *target, long value, int pe);
long long shmem_longlong_swap(long long *target, long long value, int pe);
float shmem_float_swap(float *target, float value, int pe);
double shmem_double_swap(double *target, double value, int pe);
int shmem_int_cswap(int *target, int cond, int val, int pe);
long shmem_long_cswap(long *target, long cond, long val, int pe);
long long shmem_longlong_cswap(long long *target, long long cond, long long val, int pe);
int shmem_int_fadd(int *target, int value, int pe);
long shmem_long_fadd(long *target, long value, int pe);
long long shmem_longlong_fadd(long long *target, long long value, int pe);
int shmem_int_finc(int *target, int pe);
long shmem_long_finc(long *target, int pe);
long long shmem_longlong_finc(long long *target, int pe);

/* Atomic Memory Operation Routines */
void shmem_int_add(int *target, int value, int pe);
void shmem_long_add(long *target, long value, int pe);
void shmem_longlong_add(long long *target, long long value, int pe);
void shmem_int_inc(int *target, int pe);
void shmem_long_inc(long *target, int pe);
void shmem_longlong_inc(long long *target, int pe);

/* Point-to-Point Synchronization Routines */
void shmem_short_wait(short *var, short value);
void shmem_int_wait(int *var, int value);
void shmem_long_wait(long *var, long value);
void shmem_longlong_wait(long long *var, long long value);
void shmem_wait(long *var, long value);
void shmem_short_wait_until(short *ivar, int cmp, short value);
void shmem_int_wait_until(int *ivar, int cmp, int value);
void shmem_long_wait_until(long *ivar, int cmp, long value);
void shmem_longlong_wait_until(long long *ivar, int cmp, long long value);
void shmem_wait_until(long *ivar, int cmp, long value);

/* Barrier Synchronization Routines */
void shmem_barrier_all(void);
void shmem_barrier(int PE_start, int logPE_stride, int PE_size, long *pSync);
void shmem_fence(void);
void shmem_quiet(void);

/* Reduction Routines */
/* AND to All */
void shmem_int_and_to_all(int *target, int *source, int nreduce, int PE_start, int logPE_stride, int PE_size, int *pWrk, long *pSync);
void shmem_long_and_to_all(long *target, long *source, int nreduce, int PE_start, int logPE_stride, int PE_size, long *pWrk, long *pSync);
void shmem_longlong_and_to_all(long long *target, long long *source, int nreduce, int PE_start, int logPE_stride, int PE_size, long long *pWrk, long *pSync);
void shmem_short_and_to_all(short *target, short *source, int nreduce, int PE_start, int logPE_stride, int PE_size, short *pWrk, long *pSync);
/* OR to All */
void shmem_short_or_to_all(short *target, short *source, int nreduce, int PE_start, int logPE_stride, int PE_size, short *pWrk, long *pSync);
void shmem_int_or_to_all(int *target, int *source, int nreduce, int PE_start, int logPE_stride, int PE_size, int *pWrk, long *pSync);
void shmem_long_or_to_all(long *target, long *source, int nreduce, int PE_start, int logPE_stride, int PE_size, long *pWrk, long *pSync);
void shmem_longlong_or_to_all(long long *target, long long *source, int nreduce, int PE_start, int logPE_stride, int PE_size, long long *pWrk, long *pSync);
/* XOR to All */
void shmem_short_xor_to_all(short *target, short *source, int nreduce, int PE_start, int logPE_stride, int PE_size, short *pWrk, long *pSync);
void shmem_int_xor_to_all(int *target, int *source, int nreduce, int PE_start, int logPE_stride, int PE_size, int *pWrk, long *pSync);
void shmem_long_xor_to_all(long *target, long *source, int nreduce, int PE_start, int logPE_stride, int PE_size, long *pWrk, long *pSync);
void shmem_longlong_xor_to_all(long long *target, long long *source, int nreduce, int PE_start, int logPE_stride, int PE_size, long long *pWrk, long *pSync);
/* MAX to All */
void shmem_short_max_to_all(short *target, short *source, int nreduce, int PE_start, int logPE_stride, int PE_size, short *pWrk, long *pSync);
void shmem_int_max_to_all(int *target, int *source, int nreduce, int PE_start, int logPE_stride, int PE_size, int *pWrk, long *pSync);
void shmem_long_max_to_all(long *target, long *source, int nreduce, int PE_start, int logPE_stride, int PE_size, long *pWrk, long *pSync);
void shmem_float_max_to_all(float *target, float *source, int nreduce, int PE_start, int logPE_stride, int PE_size, float *pWrk, long *pSync);
void shmem_double_max_to_all(double *target, double *source, int nreduce, int PE_start, int logPE_stride, int PE_size, double *pWrk, long *pSync);
void shmem_longlong_max_to_all(long long *target, long long *source, int nreduce, int PE_start, int logPE_stride, int PE_size, long long *pWrk, long *pSync);
void shmem_longdouble_max_to_all(long double *target, long double *source, int nreduce, int PE_start, int logPE_stride, int PE_size, long double *pWrk, long *pSync);
/* MIN to All */
void shmem_short_min_to_all(short *target, short *source, int nreduce, int PE_start, int logPE_stride, int PE_size, short *pWrk, long *pSync);
void shmem_int_min_to_all(int *target, int *source, int nreduce, int PE_start, int logPE_stride, int PE_size, int *pWrk, long *pSync);
void shmem_long_min_to_all(long *target, long *source, int nreduce, int PE_start, int logPE_stride, int PE_size, long *pWrk, long *pSync);
void shmem_float_min_to_all(float *target, float *source, int nreduce, int PE_start, int logPE_stride, int PE_size, float *pWrk, long *pSync);
void shmem_double_min_to_all(double *target, double *source, int nreduce, int PE_start, int logPE_stride, int PE_size, double *pWrk, long *pSync);
void shmem_longlong_min_to_all(long long *target, long long *source, int nreduce, int PE_start, int logPE_stride, int PE_size, long long *pWrk, long *pSync);
void shmem_longdouble_min_to_all(long double *target, long double *source, int nreduce, int PE_start, int logPE_stride, int PE_size, long double *pWrk, long *pSync);
/* SUM to All */
void shmem_short_sum_to_all(short *target, short *source, int nreduce, int PE_start, int logPE_stride, int PE_size, short *pWrk, long *pSync);
void shmem_int_sum_to_all(int *target, int *source, int nreduce, int PE_start, int logPE_stride, int PE_size, int *pWrk, long *pSync);
void shmem_long_sum_to_all(long *target, long *source, int nreduce, int PE_start, int logPE_stride, int PE_size, long *pWrk, long *pSync);
void shmem_longlong_sum_to_all(long long *target, long long *source, int nreduce, int PE_start, int logPE_stride, int PE_size, long long *pWrk, long *pSync);
void shmem_float_sum_to_all(float *target, float *source, int nreduce, int PE_start, int logPE_stride, int PE_size, float *pWrk, long *pSync);
void shmem_double_sum_to_all(double *target, double *source, int nreduce, int PE_start, int logPE_stride, int PE_size, double *pWrk, long *pSync);
void shmem_longdouble_sum_to_all(long double *target, long double *source, int nreduce, int PE_start, int logPE_stride, int PE_size, long double *pWrk, long *pSync);
void shmem_complexd_sum_to_all(double complex *target, double complex *source, int nreduce, int PE_start, int logPE_stride, int PE_size, double complex *pWrk, long *pSync);
void shmem_complexf_sum_to_all(float complex *target, float complex *source, int nreduce, int PE_start, int logPE_stride, int PE_size, float complex *pWrk, long *pSync);
/* PROD to All */
void shmem_short_prod_to_all(short *target, short *source, int nreduce, int PE_start, int logPE_stride, int PE_size, short *pWrk, long *pSync);
void shmem_int_prod_to_all(int *target, int *source, int nreduce, int PE_start, int logPE_stride, int PE_size, int *pWrk, long *pSync);
void shmem_long_prod_to_all(long *target, long *source, int nreduce, int PE_start, int logPE_stride, int PE_size, long *pWrk, long *pSync);
void shmem_longlong_prod_to_all(long long *target, long long *source, int nreduce, int PE_start, int logPE_stride, int PE_size, long long *pWrk, long *pSync);
void shmem_float_prod_to_all(float *target, float *source, int nreduce, int PE_start, int logPE_stride, int PE_size, float *pWrk, long *pSync);
void shmem_double_prod_to_all(double *target, double *source, int nreduce, int PE_start, int logPE_stride, int PE_size, double *pWrk, long *pSync);
void shmem_longdouble_prod_to_all(long double *target, long double *source, int nreduce, int PE_start, int logPE_stride, int PE_size, long double *pWrk, long *pSync);
void shmem_complexd_prod_to_all(double complex *target, double complex *source, int nreduce, int PE_start, int logPE_stride, int PE_size, double complex *pWrk, long *pSync);
void shmem_complexf_prod_to_all(float complex *target, float complex *source, int nreduce, int PE_start, int logPE_stride, int PE_size, float complex *pWrk, long *pSync);

/* Collect Routines */
void shmem_collect32(void *target, const void *source, size_t nelems, int PE_start, int logPE_stride, int PE_size, long *pSync);
void shmem_collect64(void *target, const void *source, size_t nelems, int PE_start, int logPE_stride, int PE_size, long *pSync);
void shmem_fcollect32(void *target, const void *source, size_t nelems, int PE_start, int logPE_stride, int PE_size, long *pSync);
void shmem_fcollect64(void *target, const void *source, size_t nelems, int PE_start, int logPE_stride, int PE_size, long *pSync);

/* Broadcast Routines */
void shmem_broadcast32(void *target, const void *source, size_t nelems, int PE_root, int PE_start, int logPE_stride, int PE_size, long *pSync);
void shmem_broadcast64(void *target, const void *source, size_t nelems, int PE_root, int PE_start, int logPE_stride, int PE_size, long *pSync);

/* Lock Routines */
void shmem_set_lock(long *lock);
void shmem_clear_lock(long *lock);
int shmem_test_lock(long *lock);

/* Cache Management Routines -- Unsupported */
void shmem_set_cache_inv(void);
void shmem_set_cache_line_inv(void * target);
void shmem_clear_cache_inv(void);
void shmem_clear_cache_line_inv(void * target);
void shmem_udcflush(void);
void shmem_udcflush_line(void * target); 

#ifdef __cplusplus
}
#endif 

#endif

