/* IBM_PROLOG_BEGIN_TAG                                                   */
/* This is an automatically generated prolog.                             */
/*                                                                        */
/*                                                                        */
/*                                                                        */
/* Licensed Materials - Property of IBM                                   */
/*                                                                        */
/* (C) COPYRIGHT International Business Machines Corp. 2010,2013          */
/* All Rights Reserved                                                    */
/*                                                                        */
/* US Government Users Restricted Rights - Use, duplication or            */
/* disclosure restricted by GSA ADP Schedule Contract with IBM Corp.      */
/*                                                                        */
/* IBM_PROLOG_END_TAG                                                     */
#ifndef _PRAGMA_COPYRIGHT_
#define _PRAGMA_COPYRIGHT_
#pragma comment(copyright, "@(#) 1.6.2.1 src/ppe/pnsd/nrt.h, ppe.pami, ppe_rbarlx2, rbarlx2s012a 13/03/14 14:13:02\0")
#endif /* _PRAGMA_COPYRIGHT_ */
#ifndef _NRT_H
#define _NRT_H

#include <stdint.h>
#include <netinet/in.h>

#define NRT_VERSION               1200

//
// NRT Constants
//
#define NRT_MAX_ADAPTERS_PER_TYPE    8
#define NRT_MAX_ADAPTER_NAME_LEN    32
#define NRT_MAX_DEVICENAME_SIZE       NRT_MAX_ADAPTER_NAME_LEN
#define NRT_MAX_PORTS_PER_ADAPTER    4
#define NRT_MAX_IF_NAME_LEN         16
#define NRT_MAX_JOB_NAME_LEN        64
#define NRT_MAX_PROTO_NAME_LEN      64

//
// NRT basic types
//
typedef uint32_t    nrt_task_id_t;
typedef uint32_t    nrt_endpoint_id_t;
typedef uint32_t    nrt_job_key_t;
typedef uint8_t     nrt_bool_t;
typedef uint32_t    nrt_node_number_t;
typedef uint8_t     nrt_port_id_t;
typedef uint8_t     nrt_port_status_t;
typedef uint32_t    nrt_logical_id_t;
typedef uint64_t    nrt_network_id_t;
typedef uint16_t    nrt_window_id_t;
typedef uint16_t    nrt_context_id_t;
typedef uint16_t    nrt_table_id_t;
typedef uint8_t     nrt_ib_lmc_t;
typedef uint8_t     nrt_hfi_id_t;
typedef uint16_t    nrt_hfi_isr_t;
typedef uint32_t    nrt_uid_t;
typedef uint16_t    nrt_imm_send_slot_t;
typedef uint16_t    nrt_lpar_id_t;
typedef uint16_t    nrt_cau_index_t;
typedef uint64_t    nrt_cau_resource_t;
typedef uint16_t    nrt_job_option_t;

typedef enum {
   NRT_ADAP_UNSUPPORTED=-1,
   NRT_IB=0,                // Infiniband
   NRT_HFI,                 // P7 Host Fabric Interface
   NRT_IPONLY,              // IP-Only interfaces
   NRT_HPCE,                // HPC Ethernet
   NRT_KMUX,                // Kernel Emulator of HPCE
   NRT_MAX_ADAPTER_TYPES
} nrt_adapter_t;

typedef enum { LEAVE_IN_USE, KILL } clean_option_t;

typedef enum { CKPT=0x1 } nrt_option_t;

// NRT job options bit masks
#define PREEMPT_RELEASE_RESOURCES_MASK       0x0001

typedef enum {
	PES_INIT,
	PES_JOB_RUNNING,
	PES_PREEMPTION_INPROGRESS,
	PES_JOB_PREEMPTED,
	PES_PREEMPTION_FAILED,
	PES_RESUME_INPROGRESS,
	PES_RESUME_FAILED
} preemption_state_t;

//
// Data structures for NRT_CMD_STATUS_ADAPTER
//
typedef enum { NRT_WIN_UNAVAILABLE, // Initialization stuff in progress
               NRT_WIN_INVALID,     // Not a usable window
               NRT_WIN_AVAILABLE,   // Ready for NRT load
               NRT_WIN_RESERVED,    // Window reserved, NRT not loaded
                                    //     (for POE-PMD)
               NRT_WIN_READY,       // NRT loaded
               NRT_WIN_RUNNING      // HAL has gotten NRT from us
} win_state_t;

typedef struct {
        pid_t       client_pid;      // Pid of process that loaded
        uid_t       uid;             // Uid using the window
        nrt_window_id_t    window_id;       // Window being reported
        uint16_t    bulk_transfer;   // Is this load using RDMA?
        uint32_t    rcontext_blocks; // rcontexts per window
        win_state_t state;           // Window state
        uint8_t     padding[4];
} nrt_status_t;

//
// Data structures for NRT_CMD_LOAD_TABLE
//
typedef struct {
    union {
        nrt_task_id_t               num_tasks;
        nrt_endpoint_id_t           num_endpoints;
    };
    nrt_job_key_t               job_key;
    uid_t                       uid;
    nrt_job_option_t            job_options;
    uint16_t                    pad1;
    nrt_network_id_t            network_id;
    pid_t                       pid;
    nrt_adapter_t               adapter_type;
    nrt_bool_t                  is_user_space;  // or IP
    nrt_bool_t                  is_ipv4;        // or IPv6
    nrt_context_id_t            context_id;
    nrt_table_id_t              table_id;
    char                        job_name[NRT_MAX_JOB_NAME_LEN];
    char                        protocol_name[NRT_MAX_PROTO_NAME_LEN];
    nrt_bool_t                  use_bulk_transfer;
    uint32_t                    bulk_transfer_resources;
    nrt_imm_send_slot_t         immed_send_slots_per_win;
    nrt_cau_index_t             num_cau_indexes;
} nrt_table_info_t;

typedef struct {
    union {
        nrt_task_id_t      task_id;
        nrt_endpoint_id_t  endpoint_id;
    };
    nrt_window_id_t    win_id;
    nrt_node_number_t  node_number;
    char               device_name[NRT_MAX_DEVICENAME_SIZE];
    nrt_logical_id_t   base_lid;
    nrt_port_id_t      port_id;
    nrt_ib_lmc_t       lmc;
    nrt_port_status_t  port_status;
} nrt_ib_task_info_t;

typedef struct {
    union {
        nrt_task_id_t      task_id;
        nrt_endpoint_id_t  endpoint_id;
    };
    nrt_window_id_t    win_id;
    nrt_node_number_t  node_number;
    char               device_name[NRT_MAX_DEVICENAME_SIZE];
} nrt_hpce_task_info_t;

typedef struct {
    union {
        nrt_task_id_t      task_id;
        nrt_endpoint_id_t  endpoint_id;
    };
    nrt_window_id_t    reserved;
    nrt_node_number_t  node_number;
    union {
        struct in_addr   ipv4_addr;
        struct in6_addr  ipv6_addr;
    } ip;
} nrt_ip_task_info_t;


typedef struct {
    union {
        nrt_task_id_t      task_id;
        nrt_endpoint_id_t  endpoint_id;
    };
    unsigned int       lpar_id:8;
    unsigned int       reserved:1;
    unsigned int       lid:15;          // 15-bit value: bit 0 - 0 for hfi0,
                                        //                       1 for hfi1
                                        //               bits 1-14 for ISR Address
    unsigned int       win_id:8;
} nrt_hfi_task_info_t;

//
// Data structures for NRT_CMD_QUERY_ADAPTER_INFO
//
enum { NRT_PORT_DOWN=0, NRT_PORT_UP=1, NRT_PORT_UNCONFIG=2 };

typedef struct {
    nrt_port_id_t      port_id;
    nrt_port_status_t  status;                  // 0=down, 1=up or 2=unconfig
    nrt_logical_id_t   lid;                     // lid for IB, isr_id+hfi_id for HFI
    nrt_network_id_t   network_id;              // unique for each network
    uint64_t           special;                 // adapter specific, e.g. lmc for IB,
                                                // LPAR ID for HFI
    char               if_name[NRT_MAX_IF_NAME_LEN];
    nrt_port_status_t  if_status;               // 0=down, 1=up or 2=unconfig
    in_addr_t          ipv4_addr;
    in_addr_t          ipv4_netmask;
    struct in6_addr    ipv6_addr;
    uint8_t            ipv6_prefix_len;
    uint64_t           reserved[8];
} nrt_port_info_t;

typedef struct {
    nrt_node_number_t           node_number;             // unique identifier of a node
    char                        adapter_name[NRT_MAX_ADAPTER_NAME_LEN];
    nrt_adapter_t               adapter_type;
    nrt_port_id_t               num_ports;
    nrt_port_info_t             port[NRT_MAX_PORTS_PER_ADAPTER];
    uint64_t                    rcontext_block_count;    // # of RDMA context blocks
    nrt_window_id_t             window_count;
    nrt_imm_send_slot_t   		immed_slots_avail;       // HFI only - 0 returned
                                                         // for other adapter types
    nrt_cau_index_t             cau_indexes_avail;       // HFI only - 0 returned
                                                         // for other adapter types
    unsigned char               pad[2];
    nrt_window_id_t             *window_list;
    uint64_t                    reserved[8];
} nrt_adapter_info_t;

typedef struct {
    nrt_node_number_t node_number;
    uint8_t           padding[4];
    char              device_name[NRT_MAX_DEVICENAME_SIZE];
    nrt_network_id_t  net_id;
} nrt_any_window_info_t;

//
// NRT commands
//
typedef enum {
    NRT_CMD_UNLOAD_WINDOW,
    NRT_CMD_CLEAN_WINDOW,
    NRT_CMD_QUERY_JOBS,
    NRT_CMD_LOAD_TABLE,
    NRT_CMD_UNLOAD_TABLE,
    NRT_CMD_STATUS_ADAPTER,
    NRT_CMD_PREEMPT_JOB,
    NRT_CMD_RESUME_JOB,
    NRT_CMD_QUERY_PREEMPTION_STATE,
    NRT_CMD_QUERY_ADAPTER_TYPES,
    NRT_CMD_QUERY_ADAPTER_NAMES,
    NRT_CMD_QUERY_ADAPTER_INFO,
    NRT_CMD_RESERVE_WIN_BY_DEV,
    NRT_CMD_GET_NODE_NUMBER,
    NRT_CMD_GET_TABLE_PTR,
    NRT_CMD_UPDATE_TABLE,
} nrt_cmd_type_t;

typedef struct {
    char               *adapter_name;      // IN
    nrt_adapter_t      adapter_type;       // IN
    nrt_job_key_t      job_key;            // IN
    nrt_window_id_t    window_id;          // IN
} nrt_cmd_unload_window_t;

typedef struct {
    char               *adapter_name;      // IN
    nrt_adapter_t      adapter_type;       // IN
    clean_option_t     leave_inuse_or_kill;// IN
    nrt_window_id_t    window_id;          // IN
} nrt_cmd_clean_window_t;

typedef struct {
    char               *adapter_name;      // IN
    nrt_adapter_t      adapter_type;       // IN
    nrt_job_key_t      *job_count;         // OUT
    nrt_job_key_t      **job_keys;         // OUT
} nrt_cmd_query_jobs_t;

typedef struct {
    nrt_table_info_t   *table_info;        // IN
    void               *per_task_input;    // IN - pointer to nrt_ib_task_info_t,
                                           //      nrt_hpce_task_info_t,
                                           //      or nrt_ip_task_info_t
} nrt_cmd_load_table_t;

typedef nrt_cmd_load_table_t nrt_cmd_update_table_t;

typedef struct {
    nrt_job_key_t      job_key;            // IN
    nrt_context_id_t   context_id;         // IN
    nrt_table_id_t     table_id;           // IN
} nrt_cmd_unload_table_t;

typedef struct {
    char               *adapter_name;      // IN
    nrt_adapter_t      adapter_type;       // IN
    nrt_window_id_t    *window_count;      // OUT
    nrt_status_t       **status_array;     // OUT
} nrt_cmd_status_adapter_t;

typedef struct {
    nrt_job_key_t      job_key;            // IN
    nrt_option_t       option;             // IN
    struct timeval     *timeout_val;       // IN
} nrt_cmd_preempt_job_t;

typedef struct {
    nrt_job_key_t      job_key;            // IN
    nrt_option_t       option;             // IN
    struct timeval     *timeout_val;       // IN
} nrt_cmd_resume_job_t;

typedef struct {
    nrt_job_key_t      job_key;            // IN
    preemption_state_t *state;             // OUT
} nrt_cmd_query_preemption_state_t;

typedef struct {
    unsigned int       *num_adapter_types; // OUT
    nrt_adapter_t      *adapter_types;     // OUT
} nrt_cmd_query_adapter_types_t;

typedef struct {
    nrt_adapter_t      adapter_type;       // IN
    unsigned int       *max_windows;       // OUT
    unsigned int       *num_adapter_names; // OUT
    char               adapter_names[NRT_MAX_ADAPTERS_PER_TYPE][NRT_MAX_ADAPTER_NAME_LEN];// OUT
} nrt_cmd_query_adapter_names_t;

typedef struct {
    nrt_adapter_t      adapter_type;       // IN
    char               *adapter_name;      // IN
    nrt_adapter_info_t *adapter_info;      // OUT
} nrt_cmd_query_adapter_info_t;

typedef struct {
    char                  *adapter_name;          // IN
    nrt_adapter_t         adapter_type;           // IN
    nrt_uid_t             uid;                    // IN
    nrt_job_key_t         job_key;                // IN
    nrt_window_id_t       num_windows;            // IN
    nrt_imm_send_slot_t   imm_send_slots_per_win; // IN
    nrt_any_window_info_t *reserved_win_info;     // OUT
    nrt_window_id_t       *win_list;              // OUT
} nrt_cmd_reserve_win_by_dev_t;

typedef struct {
    nrt_node_number_t     *node_number;           // OUT
} nrt_cmd_get_node_number_t;

typedef struct {
    nrt_job_key_t         job_key;                // IN
    nrt_context_id_t      context_id;             // IN
    nrt_table_id_t        table_id;               // IN
    nrt_table_info_t      *table_info;            // OUT
    void                  *task_info;             // OUT
} nrt_cmd_get_table_ptr_t;


#ifdef __cplusplus
extern "C" {
#endif // __cplusplus
int nrt_command(int version, nrt_cmd_type_t cmd_type, void *cmd);
typedef int (*nrt_command_t)(int version, nrt_cmd_type_t cmd_type, void *cmd);
#ifdef __cplusplus
}
#endif // __cplusplus

//
// NRT command return codes
//
typedef enum {
    NRT_SUCCESS,            // 0 - Success
    NRT_EINVAL,             // 1 - Invalid argument
    NRT_EPERM,              // 2 - Caller not authorized
    NRT_PNSDAPI,            // 3 - PNSD API returned an error
    NRT_EADAPTER,           // 4 - Invalid adapter
    NRT_ESYSTEM,            // 5 - System Error occurred
    NRT_EMEM,               // 6 - Memory error
    NRT_EIO,                // 7 - Adapter reports down
    NRT_NO_RDMA_AVAIL,      // 8 - No RDMA windows available
    NRT_EADAPTYPE,          // 9 - Invalid adapter type
    NRT_BAD_VERSION,        // 10- Version must match nrt_version()
    NRT_EAGAIN,             // 11- Try the call again later
    NRT_WRONG_WINDOW_STATE, // 12- Window in wrong state for this operation
    NRT_UNKNOWN_ADAPTER,    // 13- One (or more) adapters in NRT unknown
    NRT_NO_FREE_WINDOW,     // 14- For reserve any windows (none free)
    NRT_ALREADY_LOADED,     // 15- NRT with same key and task 0 config
                            //     already loaded for this device
    NRT_RDMA_CLEAN_FAILED,  // 16- task's rdma cleanup failed
    NRT_WIN_CLOSE_FAILED,   // 17- task can't close window
    NRT_WIN_OPEN_FAILED,    // 18- task can't open window
    NRT_TIMEOUT,            // 19- No response back from PNSD/job
    NRT_WRONG_PREEMPT_STATE,// 20- Invalid preemption state
    NRT_NTBL_LOAD_FAILED,   // 21- fail to load ntbl
    NRT_NTBL_UNLOAD_FAILED, // 22- fail to unload ntbl
    NRT_ERR_COMMAND_TYPE,   // 23- invalid command type
    NRT_NTBL_NOT_FOUND,     // 24- ntbl not found
    NRT_OP_NOT_VALID,       // 25- requested operation not valid for given
                            // device
    NRT_IMM_SEND_RESERVE,   // 26- error during immediate send slot reserve
                            // (HFI only)
    NRT_CAU_RESERVE,        // 27- error during CAU index reserve
                            // (HFI only)
    NRT_CAU_UNRESERVE,      // 28- error during CAU index unreserve
                            // (HFI only)
    NRT_CAU_EXCEEDED,       // 29- CAU index request exceeds available
                            // resources (HFI only)
    NRT_ERR_UPDATE_NUMTASKS,// 30- task count mismatch between target and update
                            // table
    NRT_ERR_UPDATE_JOBTYPE, // 31- job type (US vs IP) mismatch between target
                            // and update table
    NRT_ERR_UPDATE_LOCALTASK,//32- updating local tasks not permitted
    NRT_ERR_UPDATE_NOTASK,  // 33- no matching task between update and target
                            // table
    NRT_ERR_UPDATE_OPTIONS, // 34- job options mismatch between target and
                            // update table
} nrt_rc_t;

extern const unsigned int nrt_command_version;
#endif // not _NRT_H
