/* IBM_PROLOG_BEGIN_TAG                                                   */
/* This is an automatically generated prolog.                             */
/*                                                                        */
/*                                                                        */
/*                                                                        */
/* Licensed Materials - Property of IBM                                   */
/*                                                                        */
/* (C) COPYRIGHT International Business Machines Corp. 2009,2013          */
/* All Rights Reserved                                                    */
/*                                                                        */
/* US Government Users Restricted Rights - Use, duplication or            */
/* disclosure restricted by GSA ADP Schedule Contract with IBM Corp.      */
/*                                                                        */
/* IBM_PROLOG_END_TAG                                                     */
/* @(#) 1.9.1.2 src/ppe/poe/include/permapi.h, ppe.poe, comm_rbarlx2, rbarlx2s012a 13/03/14 14:13:06 */
#ifndef __PERMAPI__H__
#define __PERMAPI__H__

#include <time.h>
#include <sys/types.h>

#define RM_API_VERSION 1400

typedef void * rmhandle_t; /* Handle to resource manager */

/* These enums define the job command format as either an object or a file. */
typedef  enum {
  JOB_OBJECT=1, /* The job request is struct of type job_request_t. */
  JOB_FILE      /* The job commands are specified in a file. */
} job_request_type;

/* The definition for Boolean values */
typedef  enum {
  NO = 0,
  YES
} pe_rm_bool_t;


/* The struct to specify the job file format */
typedef struct {
  job_request_type job_format; 	/* Specifies whether the job commands are specified 
                                   as a job object or by a file. */
  void * job_command; 	/* Pointer to job commands which are specified either
                           as a job object or by a file. */
  pe_rm_bool_t retry;  /* retry == FALSE if it's an immediate job */ 
  char *scheduler;  /* The scheduler name */
} job_command_t;


/* The enum to define resource usage mode. */
typedef enum { /* add _1k9 suffix to avoid confliction with llapi.h */
  PE_RM_SHARED=1,
  PE_RM_NOT_SHARED,
  UNIQUE,
  MULTIPLE
} usage_mode_t;

/* network protocol mode - IP or User Space */
typedef enum { /* add _1k9 suffix to avoid confliction with llapi.h */
  PE_RM_IP=1,
  PE_RM_US
} network_mode_t;

#define JOB_REQ_NOT_SET -1     /* constant for values not set in job request */

/* for HFI device use only */
typedef struct {
  char *imm_send_buf;      /* immediate send buffer for a protocol */
  char *collective_groups; /* collective groups for a protocol */
} hfi_dev_attr_t;

/* The struct to define network requirements of the job. */
typedef struct {
  char *protocols ;    /* List of protocols separated by comma,
                          a string copied from 'msg_api' POE option */
  char *adapter_usage; /* List of adapter usage modes separated by comma.
                          it's a string copied from 'adapter_use' POE option. */
  char *adapter_type;  /* List of adapter names separated by comma.
                          It's a string copied from 'euidevice' */
  char *mode;          /* List of modes separated by comma, a string copied from 'euilib' */
  char *instances;     /* List of he number of instances per network, separated by comma.
                          It's a string copied from 'mp_instances' POE option */
  char *dev_type;      /* List of device type separated by comma.
                          It's a string copied from '-devtype' POE option*/
  void **dev_attr;     /* Special device characteristic, it's an array of pointers,
                          the num of pointers is aligned to num of protocols */
} network_requirements_t;



/* The job types. */ 
typedef enum {
  PARALLEL_JOB=1
} job_type_t;


/* The struct to define job resource requirements. */

typedef struct {
  job_type_t   job_type; 	/* type of job  */
  int  num_nodes; 	/* The number nodes requested for the job. */
  int  tasks_per_node;	/* The tasks per node requested. */
  int  total_tasks; 	/* The total tasks of the job. */
  char **host_names;	/* Array of names of hosts for the job */
  usage_mode_t node_usage; /* Node usage requirement */
  network_requirements_t  network_usage; /* Network usage requirement */
  pe_rm_bool_t check_pointable; /* YES indicates that job is checkpoint able. 
                             NO indicates that job is not heckpointable. */
  char *check_dir; 	/* The name of the directory for check point files. */
  char *task_affinity; 	/* Task Affinity Requirement */
  int  parallel_threads; 	/* The number of OpenMP parallel threads of each task. */
  char* pool; 		/* The job pool number or name containing the nodes on
                     which the job should be running. */
  char *save_job_file; /* MP_SAVE_RMFILE, RM will save job info into the specified file */
  char *requirements; 	/* The scheduler specific requirements, which can 
                           be passed as a string to the scheduler.*/
  char *node_topology;/*indicates that job need topology scheduling*/
  int endpoints; /* endpoints per task, the endpoints setting from MP_ENDPOINTS */
  int reserved[8]; /* reserved for future use */
} job_request_t;

/* The event types generated by the resource manager */
typedef  enum {
  JOB_ERROR_EVENT=1,    /* The event is job error message.*/
  JOB_STATE_EVENT,  /* The event to indicate that job state is changed. */
  JOB_TIMER_EVENT,  /* The event to indicate that job state has timed out. */
  JOB_CKPT_IN_PROGRESS,  /* indicates a checkpoint is in progress */
  JOB_CKPT_COMPLETE,	/* indidates a checkpoint has completed  */
} event_type_t;


/* The RM job states. */
typedef enum {
  JOB_STATE_RUNNING = 1,
  JOB_STATE_NOTRUN, 
  JOB_STATE_PREEMPTED, 
  JOB_STATE_RESUMED, 
} JobState_t;


/* The struct to define specific information of tasks assigned to run on a host. */
typedef struct {
  int  task_count; /* Number of tasks assigned to run on the host. */
  int *task_ids;   /* pointer to the Ids of the tasks assigned 
                      to run on the host. The number of tasks 
                      are specified by task_count */
  char *host_name; 	/* Pointer to host name */
  char *virtual_ip; 	/* Pointer host virtual IP address */
  char *host_address; 	/* Pointer to host IP  address */
} host_usage_t;


/* The struct to provide the job information. */
typedef struct {
  char *job_name;         /* The pointer to the job Name. */
  char *rm_id;          /* RM ID value, for checkpoint/restart */
  int procs;   		/* The total number of tasks of the job. */
  int max_instances; 	/* Maximum instances of a network 
                         for the job */
  int job_key;       	/* The job key */
  pe_rm_bool_t check_pointable; /* YES indicates that job is checkpoint able. 
                             NO indicates that job is not checkpointable. */
  char **protocol;  	/* The pointer to job protocol names. 
                         Can be a string of values.  Terminated by NULL */
  char **mode;   	/* The adapter usage mode of the job terminated by NULL. IP/US. */
  int *instance;	/* # of adapter instances on per protocol basis, terminated by -1 */
  char **devicename;  	/* The pointer to adapter device names on per protocol basis terminated by NULL.  sn_single/sn_all/en0 etc. */
  int num_network;       /* # of networks */
  int host_count; 	/* The number of Hosts assigned */
  host_usage_t *hosts; 	/* Pointer to Hosts usage assigned */
  char *rset_name; 	/* Pointer to the Resource Set name */
  char *master_virtual_ip; /* Pointer Host virtual IP address 
                              where master task runs.*/
  int mdcr_jobid; /* MetaCluster job ID */
  char *mdcr_netmask; /* MetaCluster vipserver virtual netmask */
  char *ckptdir;  /* name of checkpoint directory */
  int endpoints;  /* endpoints per task, the endpoints setting from RM */
  int reserved[8]; /* reserved for future use */
} job_info_t;


/* The struct of the events supported. */
typedef struct {
  event_type_t event; /* The place holder for event type. */
  void *event_data;   /* Refer to the actual event data type 
                         (job_info_t, msg_event_t, state_event_t) */
} job_event_t;

/* For JOB_CKPT_IN_PROGRESS, the event_data pointer would point to a structure of ckpt_start_data */
struct ckpt_start_data {
  time_t ckpt_start_time;
};

/* For JOB_CKPT_COMPLETE, the event_data pointer would point to a structure of ckpt_end_data */
struct ckpt_end_data {
  time_t ckpt_start_time;
  int ckpt_rc; /* return code from the checkpoint operation */
  pe_rm_bool_t terminate; /* yes or no to continue running the job */
  char *ckpt_msg; /* checkpoint  message */
};


/* RM connect parameter structure */
typedef struct {
  int machine_count;    	/* count of hosts/machines */
  char **machine_name;    /* Array of machine names to connect to */
  char *executable;       /* name of executable to be started */
} rm_connect_param;


/***********************************************************************
 * Function Declaration statements
 **********************************************************************/

#ifdef __cplusplus
extern "C" {
#endif
  int pe_rm_init(int *rmapi_version, rmhandle_t * resource_mgr, char *rm_id, char **error_msg);
  void pe_rm_free(rmhandle_t  *resource_mgr);
  int pe_rm_submit_job(rmhandle_t resource_mgr, job_command_t  job_cmd, char **error_msg);
  int pe_rm_get_event(rmhandle_t  resource_mgr, job_event_t **job_event, int rm_timeout, char **error_msg);
  int pe_rm_get_job_info(rmhandle_t  resource_mgr, job_info_t **job_info, char ** error_msg);
  int pe_rm_send_event(rmhandle_t  resource_mgr, job_event_t *job_event,  char ** error_msg);
  int pe_rm_free_event(rmhandle_t resource_mgr, job_event_t ** job_event );
  int pe_rm_connect(rmhandle_t resource_mgr, rm_connect_param *connect_param, int *rm_sockfds, int rm_timeout, char **error_msg);
#ifdef __cplusplus
}
#endif /* __cpluplus */

#endif
