/* IBM_PROLOG_BEGIN_TAG                                                   */
/* This is an automatically generated prolog.                             */
/*                                                                        */
/*                                                                        */
/*                                                                        */
/* Licensed Materials - Property of IBM                                   */
/*                                                                        */
/* (C) COPYRIGHT International Business Machines Corp. 2011,2012          */
/* All Rights Reserved                                                    */
/*                                                                        */
/* US Government Users Restricted Rights - Use, duplication or            */
/* disclosure restricted by GSA ADP Schedule Contract with IBM Corp.      */
/*                                                                        */
/* IBM_PROLOG_END_TAG                                                     */
/* list_algs.c -- Call PAMI to list the available collective algorithms */
/* This is an MPI program that must be run under the control of POE */
/* Use POE environment variables (MP_PROCS, MP_SHARED_MEMORY, etc.) to
   define the geometry for which the PAMI collective algorithms are to
   be listed. */
/* The list will be those algorithms corresponding to MPI_COMM_WORLD */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <mpi.h>
#include <pami.h>

/* These names must match the order in pami.h */

char *mpi_names[] = {
    "MPI_BROADCAST",
    "MPI_ALLREDUCE",
    "MPI_REDUCE",
    "MPI_ALLGATHER",
    "UNDEFINED",
    "MPI_ALLGATHERV",
    "MPI_SCATTER",
    "UNDEFINED",
    "MPI_SCATTERV",
    "MPI_GATHER",
    "UNDEFINED",
    "MPI_GATHERV",
    "MPI_BARRIER",
    "MPI_ALLTOALL",
    "UNDEFINED",
    "MPI_ALLTOALLV",
    "MPI_SCAN or MPI_EXSCAN",
    "MPI_REDUCE_SCATTER",
    "UNDEFINED",
    "UNDEFINED",
    "UNDEFINED",
    "UNDEFINED" };

main(int argc, char **argv) {
unsigned long long client, context;
size_t lengths[2];
void *geoptr;
char *cp;
int rc;
int lmax;
int i;
pami_algorithm_t *alg0, *alg1;
pami_metadata_t *metad0, *metad1;
int xfer_type;
int task_count;
int my_task = 0;

/* For any argument, print help message */
cp = getenv("MP_CHILD");
if (cp) my_task = atoi(cp);
if (argc > 1) {
  if (my_task == 0) {
	printf("list_algs is an MPI program to list the PAMI collective\n");
	printf(" algorithms available to the task geometry defined by the\n");
	printf(" host list, shared memory, and number of tasks specified by POE\n");
    printf(" The list will be those algorithm names corresponding to\n");
	printf(" MPI_COMM_WORLD for this job.\n");
	}
  exit(1);
  }
	
	client = 0; 
	context = 0;
    /* Initialize MPI */
	MPI_Init(0,0);
    MPI_Comm_rank(MPI_COMM_WORLD,&my_task);
    MPI_Comm_size(MPI_COMM_WORLD,&task_count);
	/* printf("MPI Initialized\n"); */
	/* The environment variables MP_S_PAMI_CLIENT and MP_S_PAMI_CONTEXT
	   are set by MPI_Init */
	cp = getenv("MP_S_PAMI_CLIENT");
	if (cp) client = atoll(cp);
	else {
		fprintf(stderr,"MP_S_PAMI_CLIENT environment variable not set\n");
		exit(1);
		}
	cp = getenv("MP_S_PAMI_CONTEXT");
	if (cp) context = atoll(cp);
	else {
		fprintf(stderr,"MP_S_PAMI_CONTEXT environment variable not set\n");
		exit(1);
		}

	/* Print an identification of the geometry */
	if (my_task == 0) {
          char *dup = "0:";
	  cp = getenv("MP_COMMON_TASKS");
          if (cp) dup = strdup(cp);
	  dup = strtok(dup,":");
	  printf("List MPI Collective algorithms provided by PAMI\n");
	  printf("Total number of tasks = %d\n", task_count);
	  printf("Number of tasks sharing this node = %d\n", 1 + atoi(dup));

    /* Call PAMI to define the geometry for this MPI_COMM_WORLD */
	rc = PAMI_Geometry_world((void*)client, &geoptr);	

    /* PAMI_XFER_COUNT is defined in the include file pami.h */
    /* For each MPI collective in PAMI's list, there are two
		lists of algorithms: A) those that are always usable;
		B) those that are usable only in certain cases.  
		Both lists are printed. */
	/* Note that the algorithms available from PAMI depend upon the
	    geometry defined, which should correspond to the geometry to
	    be used for the job */
	
	for (xfer_type=0;xfer_type<PAMI_XFER_COUNT;xfer_type++) {
    /* PAMI has some algorithms for callers other than MPI */
        if (!strcmp(mpi_names[xfer_type],"UNDEFINED")) continue;
	rc = PAMI_Geometry_algorithms_num(geoptr, xfer_type,
		&(lengths[0]));
	lmax = (lengths[0]>lengths[1]?lengths[0]:lengths[1]);
	/* allocate enough space for the lists */
	alg0 = (pami_algorithm_t *)malloc(lmax * sizeof(pami_algorithm_t));
	metad0 = (pami_metadata_t *)malloc(lmax * sizeof(pami_metadata_t));
	alg1 = (pami_algorithm_t *)malloc(lmax * sizeof(pami_algorithm_t));
	metad1 = (pami_metadata_t *)malloc(lmax * sizeof(pami_metadata_t));

	rc = PAMI_Geometry_algorithms_query(geoptr, xfer_type,
		alg0, metad0, lengths[0], alg1, metad1, lengths[1]);
    printf("\n----- Transfer Type %s -----\n", mpi_names[xfer_type] );
 
	/* print the list of algorithms that always work */
	for (i=0; i<lengths[0]; i++) {
		printf("Always\t%s\n", metad0[i].name);
		}
	/* print the list of algorithms that need further checking */
	for (i=0; i<lengths[1]; i++) {
		printf("Check\t%s\n", metad1[i].name);
		}
	free(alg0);
	free(alg1);
	free(metad0);
	free(metad1);
	} /* end of list for MPI collective */
    } /* end of work for task 0 */

	/* Terminate MPI and exit the program */
		
	MPI_Finalize();
	exit(0);
	}
