/* IBM_PROLOG_BEGIN_TAG                                                   */
/* This is an automatically generated prolog.                             */
/*                                                                        */
/*                                                                        */
/*                                                                        */
/* Licensed Materials - Property of IBM                                   */
/*                                                                        */
/* (C) COPYRIGHT International Business Machines Corp. 1994,2012          */
/* All Rights Reserved                                                    */
/*                                                                        */
/* US Government Users Restricted Rights - Use, duplication or            */
/* disclosure restricted by GSA ADP Schedule Contract with IBM Corp.      */
/*                                                                        */
/* IBM_PROLOG_END_TAG                                                     */
/*
 *   File name: mcp.c
 *
 *   History: 
 *      11/21/97 SVC M01 Port to OS/390 Unix System Services (D35248)
 *      03/27/98 AJI M02 changes to #define BLOCK_SIZE (D38675)
 *      05/21/98 AJI M03 changes to dynamic BLOCK_SIZE (D39519)
 *      06/17/98 PFC M04 fix labels on messages added by defects 38675 and
 *                       39519 (D41184)
 *      06/18/98 SVC M05 stdlib.h needed for getenv() (D40375)
 *      02/03/99 PFC M06 check for DFS f/s when checking space (D46200)
 *      03/11/99 SVC M07 OS/390: move to mohonk, undo M06 (no DFS for OS/390) (D47705)
 *      04/20/05 ECL M08 Linux port (F118848/D120380)
 *
 *   Origins:
 *                   Copyright IBM Corporation 1994,2002
 *
 *                         All Rights Reserved
 *
 *   Permission to use, copy, modify, and distribute this software and its
 *   documentation for any purpose and without fee is hereby granted,
 *   provided that the above copyright notice appear in all copies and that
 *   both that copyright notice and this permission notice appear in
 *   supporting documentation, and that the name of IBM not be
 *   used in advertising or publicity pertaining to distribution of the
 *   software without specific, written prior permission.
 *
 *   Information: mcp is a command which allows the user to propogate a copy
 *   of a file to multiple nodes on an IBM P-Series system. The file
 *   must initially reside (or be NFS mounted) on at least one node. The format
 *   of the command is as follows: "mcp infile [outfile] [POE options]" where
 *   infile is the name of the file to be copied and outfile is the name of the
 *   the file or directory to be copied to. If outfile is not given, the file will be
 *   placed in its current directory on each node. mcp is a POE program,
 *   therefore all POE options are available to the user. POE options
 *   are set by using either command line flags or environment variables.
 *   The number of nodes to copy the file to (-procs), and the message
 *   passing protocol used to copy the file (-euilib) are the POE
 *   options which may be of most interest to the user. See the 
 *   Parallel Environment for AIX Operation and Use guide for 
 *   information on POE options.
 *   
 *   A default of 100K data block size is used for copying the data.  This can be
 *   changed by setting the enviornment variable MP_BLKSIZE between 1 and 8000000.
 *
 *
 *   Note: A POE job will run faster if a copy of the job resides on each
 *   node. It is suggested that you use mcp to copy your executable to 
 *   a filesystem such as /tmp (which resides on each node). 
 *
 *   return codes:  -1 : incorrect usage                    
 *                  -2 : error opening input file           
 *                  -3 : error opening "to" file on originating node
 *                  -4 : error writing data to "to" file on originating node
 *                  -5 : no room on remote node's filesystem                 
 *                  -6 : error opening file on remote node
 *                  -7 : error writing data on remote node
 *                  -8 : error renaming temp file to filename
 *                  -9 : input file is empty
 *                 -10 : invalid block size
 *                 -11 : error allocating storage
 *
 */
#pragma comment(copyright,"@(#) 5765-F83 (C) Copyright IBM Corp. 1994,2002\0")
#pragma comment(user,"@(#) 1.11.1.13 src/ppe/poe/samples/mpi/mcp.c, ppe.poe.samples, ppe_rbarlx2, rbarlx21216b 12/04/23 13:26:33")

#include <stdio.h>
#include <stdlib.h>      /* M05 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/statvfs.h>

#ifndef __linux   /* M08: vnode.h not available on Linux */
  #include <sys/vnode.h>
#endif        

#include <mpi.h>

#define MAX_BLOCK_SIZE 8000000  /* M02 maximum block size user can request */

main(argc,argv)
int argc;
char *argv[];
{
  char *in_file;               /* name of file to be copied */
  char *out_file;              /* optional output file name */
  FILE *fp;                    /* file pointer for input file */
  FILE *fp_out;                /* file pointer for file being written */
  struct stat statb;           /* structure needed for stat() function */
  struct statvfs stata;        /* structure needed for statvfs() function */

#ifndef __linux                  /* M08: VDIR & st_type not available on Linux */
  #define STATB_STTYPE_IS_VDIR statb.st_type == VDIR
#else
  #define STATB_STTYPE_IS_VDIR S_ISDIR(statb.st_mode)
#endif
  int i,j,k;                   /* general purpose integers */
  int opened_file;             /* indicates if out file is open for writing */

  struct file_def {            /* file data broadcast to each node */
    int  name_size;            /* size of file name */
    long  file_size;            /* size of the file */
    int  content_size;         /* size of file data in this broadcast */
    char name[FILENAME_MAX];   /* file name */
    mode_t file_mode;          /* file mode (permisssions etc...) */
  } file_definition;
 
  char temp_filename[FILENAME_MAX]; /* temporary filename */
  char *in_name; /* input file name */
  char *env_block_ptr;         /* M02 pointer to env var to get MP_BLKSIZE value */
  char *contentps;             /* M03 pointer to send file data */
  char *contentpr;             /* M03 pointer to recieve file data */
  char *raw_buffer;            /* M03 pointer to send and recieve data buffer */
  struct file_def  *fdp;       /* M03 */
  long int blocksize = 100000; /* M02 default blocksize is used if MP_BLKSIZE not set */


  int number_of_tasks;         /* # nodes (including sender) executing */
  int my_task_number;          /* task number of this node */
  int msg_rc;                  /* return code from broadcast */
  int buf[4],elements = 4, query_types=3, ALLGRP; /* data for mpc_task_query */

MPI_Init(&argc,&argv); /* initialize MPI environment */
MPI_Comm_rank(MPI_COMM_WORLD, &my_task_number); /* get task number */
MPI_Comm_size(MPI_COMM_WORLD,&number_of_tasks); /* get number of tasks */

/***********************************************/
/* if this is the node which contains the file */
/* to be copied, check for correct argument    */
/* usage, then open the file.                  */
/***********************************************/

  if(my_task_number == 0) {

    if(argc < 2 || argc > 3) {    /* should be 1-2 arguments to this prog */
      fprintf(stderr,"usage: mcp infile [outfile]\n"); /* if not, put msg */
      MPI_Abort(MPI_COMM_WORLD,-1); /* stop all other tasks */
    }

/***********************************************/
/* if input and output files specified, get    */
/* name and len of output file. if just input  */
/* file specified, get its name and length.    */
/***********************************************/

    file_definition.name_size = strlen(argv[argc-1]); /* get len file */
    strcpy(file_definition.name,argv[argc-1]);    /* get name of file */
    in_file = (char *) malloc(strlen(argv[1])+1);   /*stg for input file name*/
    strcpy(in_file,argv[1]);                      /* get input file name */

    /* If destination output file is a directory, concatenate the input file 
       name to it, following a slash (/).        */

    stat(file_definition.name,&statb);

    if (STATB_STTYPE_IS_VDIR ) { 
       strcat(file_definition.name,"/");
       in_name = (char *)strrchr(in_file,'/');  /* Strip off path from input file name */
       in_name ? strcat(file_definition.name,++in_name) : 
                 strcat(file_definition.name,in_file);  
                                          /* Join directory name and file name, after 
                                              stripping off input path name */
    }

/***********************************************/
/* if no information is available about the    */
/* file or the file cannot be opened...        */
/***********************************************/

    if(stat(in_file,&statb) == -1 || ! (fp=fopen(in_file,"r"))) {
      fprintf(stderr,"mcp: Could not open file %s for reading\n",in_file);
      MPI_Abort(MPI_COMM_WORLD,-2); /* stop all other tasks */
    }

    if (statb.st_size == 0) {     /* File is empty....issue msg */
      fprintf(stderr,"mcp: File %s is empty\n",in_file);
      MPI_Abort(MPI_COMM_WORLD,-9); /* stop all other tasks */
    }

    file_definition.file_mode = statb.st_mode;  /* get file mode */
    file_definition.file_size = statb.st_size;  /* get file len from struct */
  }                                             /* end of "if this is task 0" */

/***********************************************/
/* If this is the task with the originating    */
/* file and a "to" file was specified, we need */
/* to copy the file to the "to" file on this   */
/* node so we'll open up a "to" file.          */
/***********************************************/

  opened_file = 0;                       /* file not yet opened for this task */
  if((my_task_number == 0) && (argc == 3)) { /*if task 0 + there is "to" file*/
    if( !(fp_out = fopen(file_definition.name,"w")) ) { /* open "to" file */
      fprintf(stderr," mcp: Failed on open of %s on node %d\n",
              file_definition.name,my_task_number);
      MPI_Abort(MPI_COMM_WORLD,-3); /* stop all other tasks */
    }
    opened_file = 1;                      /* "to" file on originating opened */
  }

/***********************************************/
/* Check to see if env varible is set thru cmd */
/* line. If the environment variable was set   */
/* get the value of it. Check to see if BLOCK  */
/* size is with in limits, if so allocate the  */
/* the memory based on the BLOCK SIZE.  M02    */
/***********************************************/

  if ((env_block_ptr = getenv("MP_BLKSIZE")) != NULL) {  /* M02 */
     blocksize = atoi(env_block_ptr);  /* M02 */
   }     /* M02 */
  if ((blocksize <= 0) || (blocksize > MAX_BLOCK_SIZE)) {    /* M02 */
     fprintf(stderr, "mcp: blocksize has to be between 1 and %ld\n", MAX_BLOCK_SIZE);
                                                                    /* M02 */
                     /* M04 added function label to message */
     MPI_Abort(MPI_COMM_WORLD,-10); /* M02 stop all other tasks */
   }    /* M02 */

/* Allocate data area based on block size and file definition to broadcast data  */

  if ((raw_buffer = (char *) malloc(sizeof(file_definition)+blocksize)) == NULL) {
                              /* M03  raw_buffer will be used to broadcast data */
     fprintf(stderr,"mcp: error allocating storage for send and recieve \n"); /* M03 */
                     /* M04 added function label to message */
     MPI_Abort(MPI_COMM_WORLD,-11); /* M03 stop all other tasks */
   }    /* M03 */

  contentps = raw_buffer + sizeof(file_definition); /* M03 pointer to content area */

/***********************************************/
/* This is the main loop for broadcasting the  */
/* file. If this is task 0, read a block of    */
/* the file. If this is any other task, open   */
/* the file for writing (if it hasn't          */
/* previously been opened) and write the block.*/
/* Continue to do the above until the entire   */
/* file has been sent.                         */
/***********************************************/
    
  do {                                   /*cont reading and writing until eof*/
    if (my_task_number == 0) {
      if(statb.st_size > blocksize) {    /* M02 if # bytes left in file > block */
        k =blocksize;                    /* M02 write amount = 1 block */
      }
      else {                             /* if # bytes left < or = block */
        k = statb.st_size;               /* write amount = amount left */
      }
      statb.st_size -= k;                /* decrement amount left to write */
      if( !fread(contentps,sizeof(char),k,fp)) { /* M03 read piece */
        fprintf(stderr,"mcp Attention: fread of %s may have failed\n",
                file_definition.name);
      }
      file_definition.content_size = k;  /* # bytes read in structure sent */
 
/************************************************/
/* if there is a "to" file, write the file to   */
/* the originating node.                        */
/************************************************/

      if(argc == 3) {                    /* if there is a "to" file */
        if((j=fwrite(contentps,sizeof(char),k,fp_out))!=k) {  /* M03 */
          fprintf(stderr," mcp: Failed on write of %s on node %d\n",
                  file_definition.name,my_task_number); /* write failure */
          fprintf(stderr, "mcp: Wrote %d bytes while writing %d\n",j,k);
          MPI_Abort(MPI_COMM_WORLD,-4); /* stop all other tasks */
        }
      }
    }                                    /* end if task = 0 */

/*******************************/
/* broadcast out the file name */
/*******************************/

    memcpy(raw_buffer, (char *) &file_definition, sizeof(file_definition));
                               /* M03  prepare the raw_buffer to broadcast */  
    msg_rc = MPI_Bcast(raw_buffer, sizeof(file_definition)+blocksize,MPI_CHAR,0,
             MPI_COMM_WORLD); /* M03 broadcast file definition and contents at once */  

    if(my_task_number > 0) {             /* if this is not task 0 */

/************************************************/
/* Verify that there is space for the file.     */
/* Divide the file size by the block size of    */  
/* the file system to find out how many blocks  */
/* the file will need. If that value is > the   */
/* number of free blocks, then stop processing  */
/* this task and return.                        */
/************************************************/

      fdp = (struct file_def *)raw_buffer;   /* M03 */
      contentpr = raw_buffer + sizeof(file_definition); /* M03 */
 
      if(!(statvfs(fdp->name, &stata))) {  /* M03 */

      /* M06 Check file system space only if file sys type is not DFS */

#ifndef __linux   /* M07/M08: No DFS support on Linux */
        if (strncmp(stata.f_basetype,"dfs",3) != 0)    /* M06 */
#endif          /* M07/M08 */
          if (stata.f_bavail<((unsigned long)fdp->file_size/stata.f_frsize+1)) { /* M03 */

            printf("mcp:no room in filesystem to copy %s for task %d.\n", 
                    fdp->name, my_task_number); /* M03 */
            MPI_Abort(MPI_COMM_WORLD,-5); /* stop all other tasks */
          }
      }
      if(opened_file == 0) {             /*if output file has not been opened*/
 
        sprintf(temp_filename,"%s.%d\0",fdp->name, getpid());  /* M03 */
        if( !(fp_out = fopen(temp_filename,"w")) ) {    /* open it */
          fprintf(stderr," mcp: Failed opening %s on node %d\n",
                  temp_filename,my_task_number);
          MPI_Abort(MPI_COMM_WORLD,-6); /* stop all other tasks */
        }
        opened_file = 1;                 /* indicate file has been opened */
        statb.st_size = fdp->file_size; /* M03 # bytes expected to write*/
      }
      k = fdp->content_size;  /* M03 write amount = amount sent */
      if((j=fwrite(contentpr,sizeof(char),k,fp_out))!=k) { /* M03 */
        fprintf(stderr," mcp: Failed writing %s on node %d\n",
                temp_filename,my_task_number); /* write failure */
        fprintf(stderr, "mcp: Wrote %d bytes while writing %d\n",j,k);
        MPI_Abort(MPI_COMM_WORLD,-7); /* stop all other tasks */
      }
      statb.st_size -= k;      /* decrement total by amount written this time */
    }                                /* my_task_number > 0 */

  } while (statb.st_size > 0); /* cont above until entire file is read/written*/

  if(my_task_number == 0) { /* if this is task 0 */
    fclose(fp);                      /* close the file we've been reading */
    if(argc == 3) {                  /* if there is a "to" file */
      fclose(fp_out);                /* close the file we've been writing */
      chmod(file_definition.name,file_definition.file_mode); /* set file mode */
    }
    free(in_file);
  }
  else {                             /* if this isn't task 0 */
    fclose(fp_out);                  /* close the file we've been writing */
    chmod(temp_filename,fdp->file_mode); /*  M03 set file mode */
    if(rename(temp_filename,fdp->name) != 0) { /* M03 rename temp file */
      fprintf(stderr, "mcp: Error renaming temp file %s to %s\n",
              temp_filename, fdp->name);  /* M03 */
      MPI_Abort(MPI_COMM_WORLD,-8); /* stop all other tasks */
    }
  }
free(raw_buffer); /* M03 */
MPI_Barrier(MPI_COMM_WORLD);  
MPI_Finalize();
}
